/* -----------------------------------------------------------------------------
 * Copyright (c) 2013-2015 ARM Ltd.
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software. Permission is granted to anyone to use this
 * software for any purpose, including commercial applications, and to alter
 * it and redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software in
 *    a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 *
 *
 * $Date:        10. June 2015
 * $Revision:    V1.00
 *
 * Driver:       Driver_USBD0
 * Project:      USB High-Speed Device Driver for Atmel SAMV71
 * --------------------------------------------------------------------------
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 *
 *   Configuration Setting                Value
 *   ---------------------                -----
 *   Connect to hardware via Driver_USBD# = 0
 * -------------------------------------------------------------------------- */

/* History:
 *  Version 1.00
 *    Initial release
 */
 
#include <stdint.h>
#include <string.h>
#include "Driver_USBD.h"
#include "USBD_SAMV71.h"

// USBD Driver *****************************************************************

#define ARM_USBD_DRV_VERSION ARM_DRIVER_VERSION_MAJOR_MINOR(1,0) // USBD driver version

// Driver Version
static const ARM_DRIVER_VERSION usbd_driver_version = { ARM_USBD_API_VERSION, ARM_USBD_DRV_VERSION };

// Driver Capabilities
static const ARM_USBD_CAPABILITIES usbd_driver_capabilities = {
#ifdef ARM_USBD_VBUS_DETECT
  1U,  // vbus_detection
  1U,  // event_vbus_on
  1U,  // event_vbus_off
#else
  0U,  // vbus_detection
  0U,  // event_vbus_on
  0U,  // event_vbus_off
#endif
};

// Number of Endpoints
#ifndef USBD_EP_NUM
#define USBD_EP_NUM             CHIP_USB_NUMENDPOINTS
#endif

// USB Device Endpoint flags
#define USBD_EP_FLAG_CONFIGURED  (1U)
#define USBD_EP_FLAG_BUSY        (1U << 1)

// Endpoint number and index
#define EP_NUM(ep_addr)           (ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK)
#define EP_ID(ep_addr)            (EP_NUM(ep_addr) * 2U + ((ep_addr >> 7U) & 1U))

// Endpoint structure
typedef struct {
  uint32_t  bufferIndex;
} ENDPOINT;

// Local variables and structures
static ARM_USBD_SignalDeviceEvent_t   cbDeviceEvent;
static ARM_USBD_SignalEndpointEvent_t cbEndpointEvent;
static ARM_USBD_STATE usbd_state = {0};
static ENDPOINT ep[(USBD_EP_NUM + 1U)*2U];
static uint8_t drv_state = 0;
static uint8_t setup_buf[8];
volatile static uint8_t setup_flag = 0U;


// USB DEVICE EVENT HANDLERS

void USBD_SuspendHandler(void)
{
  /* Suspend HW interface */
  USBD_HAL_Suspend();
  usbd_state.active = false;
  /* Send event */
  cbDeviceEvent(ARM_USBD_EVENT_SUSPEND);
}

void USBD_ResumeHandler(void)
{
  /* Active the device */
  USBD_HAL_Activate();
  usbd_state.active = true;
  /* Send event */
  cbDeviceEvent(ARM_USBD_EVENT_RESUME);
}

void USBD_ResetHandler()
{
  /* Active the USB HW */
  USBD_HAL_Activate();
  /* Only EP0 enabled */
  USBD_HAL_ResetEPs(0xFFFFFFFF, USBD_STATUS_RESET, 0);
  USBD_HAL_ConfigureEP(0);
  /* Send event */
  cbDeviceEvent(ARM_USBD_EVENT_RESET);
}

void USBD_RequestHandler(uint8_t bEndpoint,
                         const USBGenericRequest* pRequest)
{
  memcpy(setup_buf, pRequest, 8U);
  setup_flag = 1U;
  cbEndpointEvent(bEndpoint, ARM_USBD_EVENT_SETUP); 
}

void USBD_TransferCallback(uint8_t ep_addr, uint8_t status, uint32_t transferred, uint32_t remaining) {  
  /* Check status */
  if (status != USBD_STATUS_SUCCESS) return;
    
  /* Increment buffer index */
  ep[EP_ID(ep_addr)].bufferIndex += transferred;    
  
  /* Send event */
  if (ep_addr&0x80) {
    cbEndpointEvent(ep_addr, ARM_USBD_EVENT_IN);
  } else {
    cbEndpointEvent(ep_addr, ARM_USBD_EVENT_OUT);  
  }
}


// USB DEVICE DRIVER FUNCTIONS

/**
  \fn          ARM_DRIVER_VERSION USBD_GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRIVER_VERSION
*/
static ARM_DRIVER_VERSION USBD_GetVersion (void) { return usbd_driver_version; }

/**
  \fn          ARM_USBD_CAPABILITIES USBD_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_USBD_CAPABILITIES
*/
static ARM_USBD_CAPABILITIES USBD_GetCapabilities (void) { return usbd_driver_capabilities; }

/**
  \fn          int32_t USBD_Initialize (ARM_USBD_SignalDeviceEvent_t   cb_device_event,
                                        ARM_USBD_SignalEndpointEvent_t cb_endpoint_event)
  \brief       Initialize USB Device Interface.
  \param[in]   cb_device_event    Pointer to \ref ARM_USBD_SignalDeviceEvent
  \param[in]   cb_endpoint_event  Pointer to \ref ARM_USBD_SignalEndpointEvent
  \return      \ref execution_status
*/
static int32_t USBD_Initialize (ARM_USBD_SignalDeviceEvent_t   cb_device_event,
                                ARM_USBD_SignalEndpointEvent_t cb_endpoint_event) {
                         
  if (drv_state & USBD_DRIVER_INITIALIZED) { return ARM_DRIVER_OK; }
  if (drv_state)                           { return ARM_DRIVER_ERROR; }                                                                

  cbDeviceEvent = cb_device_event;
  cbEndpointEvent = cb_endpoint_event;
    
  drv_state = USBD_DRIVER_INITIALIZED;
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_Uninitialize (void)
  \brief       De-initialize USB Device Interface.
  \return      \ref execution_status
*/
static int32_t USBD_Uninitialize (void) {

  drv_state = 0U;
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_PowerControl (ARM_POWER_STATE state)
  \brief       Control USB Device Interface Power.
  \param[in]   state  Power state
  \return      \ref execution_status
*/
static int32_t USBD_PowerControl (ARM_POWER_STATE state) {

  switch (state) {
    case ARM_POWER_OFF:      
      /* Disable IRQ */
      NVIC_DisableIRQ(USBHS_IRQn) ; 
      /* Disable peripheral clock for USBHS */
      PMC_DisablePeripheral(ID_USBHS);
      
      drv_state  &= ~USBD_DRIVER_POWERED;
      break;

    case ARM_POWER_FULL:
      if (!(drv_state & USBD_DRIVER_INITIALIZED)) { return ARM_DRIVER_ERROR; }
      if (drv_state & USBD_DRIVER_POWERED) { return ARM_DRIVER_OK; }
      
      /* UTMI parallel mode, High/Full/Low Speed */
      /* UUSBCK not used in this configuration (High Speed) */
      PMC->PMC_SCDR = PMC_SCDR_USBCLK;
      /* USB clock register: USB Clock Input is UTMI PLL */
      PMC->PMC_USB = PMC_USB_USBS;
      /* Enable peripheral clock for USBHS */
      PMC_EnablePeripheral(ID_USBHS);    
      USBHS->USBHS_CTRL = USBHS_CTRL_UIMOD_DEVICE;
      /* Enable PLL 480 MHz */
      PMC->CKGR_UCKR = CKGR_UCKR_UPLLEN | CKGR_UCKR_UPLLCOUNT(0xF);
      /* Wait that PLL is considered locked by the PMC */
      while( !(PMC->PMC_SR & PMC_SR_LOCKU) ); 
      
      USBD_HAL_Init();
      
      usbd_state.speed = ARM_USB_SPEED_HIGH; 
      drv_state |= USBD_DRIVER_POWERED;
      break;
    
    default:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
  }
      
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_DeviceConnect (void)
  \brief       Connect USB Device.
  \return      \ref execution_status
*/
static int32_t USBD_DeviceConnect (void) {
  if (!(drv_state & USBD_DRIVER_POWERED) ) { return ARM_DRIVER_ERROR; }
  if (  drv_state & USBD_DRIVER_CONNECTED) { return ARM_DRIVER_OK; }
  
  USBD_HAL_Connect();
  usbd_state.active = true;
  drv_state |= USBD_DRIVER_CONNECTED;
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_DeviceDisconnect (void)
  \brief       Disconnect USB Device.
  \return      \ref execution_status
*/
static int32_t USBD_DeviceDisconnect (void) {
  if (!(drv_state & USBD_DRIVER_POWERED)  ) { return ARM_DRIVER_ERROR; }
  if (!(drv_state & USBD_DRIVER_CONNECTED)) { return ARM_DRIVER_OK; }
  
  USBD_HAL_Disconnect();
  usbd_state.active = false;
  drv_state &= ~USBD_DRIVER_CONNECTED;
  return ARM_DRIVER_OK;
}

/**
  \fn          ARM_USBD_STATE USBD_DeviceGetState (void)
  \brief       Get current USB Device State.
  \return      Device State \ref ARM_USBD_STATE
*/
static ARM_USBD_STATE USBD_DeviceGetState (void) {
  return usbd_state;
}

/**
  \fn          int32_t USBD_DeviceRemoteWakeup (void)
  \brief       Trigger USB Remote Wakeup.
  \return      \ref execution_status
*/
static int32_t USBD_DeviceRemoteWakeup (void) {
  if (!(drv_state & USBD_DRIVER_CONNECTED)) { return ARM_DRIVER_ERROR; }
    
  USBD_HAL_RemoteWakeUp();
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_DeviceSetAddress (uint8_t dev_addr)
  \brief       Set USB Device Address.
  \param[in]   dev_addr  Device Address
  \return      \ref execution_status
*/
static int32_t USBD_DeviceSetAddress (uint8_t dev_addr) {
  if (!(drv_state & USBD_DRIVER_CONNECTED)) { return ARM_DRIVER_ERROR; }
  
  USBD_HAL_SetAddress(dev_addr);  
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_ReadSetupPacket (uint8_t *setup)
  \brief       Read setup packet received over Control Endpoint.
  \param[out]  setup  Pointer to buffer for setup packet
  \return      \ref execution_status
*/
static int32_t USBD_ReadSetupPacket (uint8_t *setup) {
  if (!(drv_state & USBD_DRIVER_CONNECTED)) { return ARM_DRIVER_ERROR; }
  if (!setup_flag)                          { return ARM_DRIVER_ERROR; }

  setup_flag = 0U;  
  memcpy(setup, setup_buf, 8U);
  
  if (setup_flag) {
    // New setup packet was received while this was being read
    return ARM_DRIVER_ERROR;
  }
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_EndpointConfigure (uint8_t  ep_addr,
                                               uint8_t  ep_type,
                                               uint16_t ep_max_packet_size)
  \brief       Configure USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[in]   ep_type  Endpoint Type (ARM_USB_ENDPOINT_xxx)
  \param[in]   ep_max_packet_size Endpoint Maximum Packet Size
  \return      \ref execution_status
*/
static int32_t USBD_EndpointConfigure (uint8_t  ep_addr,
                                       uint8_t  ep_type,
                                       uint16_t ep_max_packet_size) {                                     
  USBEndpointDescriptor descriptor;                                    

  if (!(drv_state & USBD_DRIVER_CONNECTED)) { return ARM_DRIVER_ERROR; }                                        
  if (EP_ID(ep_addr) > USBD_EP_NUM) { return ARM_DRIVER_ERROR; }
    
  // Set Endpoint resources
  if (EP_NUM(ep_addr)==0) {
    USBD_HAL_ConfigureEP(0); 
  } else { 
    descriptor.bEndpointAddress = ep_addr;
    descriptor.wMaxPacketSize = ep_max_packet_size & ARM_USB_ENDPOINT_MAX_PACKET_SIZE_MASK;
    descriptor.bmAttributes = ep_type; 
    USBD_HAL_ConfigureEP(&descriptor);     
  }                                                
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_EndpointUnconfigure (uint8_t ep_addr)
  \brief       Unconfigure USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \return      \ref execution_status
*/
static int32_t USBD_EndpointUnconfigure (uint8_t ep_addr) {
  if (!(drv_state & USBD_DRIVER_CONNECTED)) { return ARM_DRIVER_ERROR; }
    
  USBD_HAL_ResetEPs(1<<EP_NUM(ep_addr), USBD_STATUS_RESET, 0);
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_EndpointStall (uint8_t ep_addr, bool stall)
  \brief       Set/Clear Stall for USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[in]   stall  Operation
                - \b false Clear
                - \b true Set
  \return      \ref execution_status
*/
static int32_t USBD_EndpointStall (uint8_t ep_addr, bool stall) {
  int32_t result;
  
  if (!(drv_state & USBD_DRIVER_CONNECTED)) { return ARM_DRIVER_ERROR; }

  if (EP_NUM(ep_addr)==0) {
    if (stall) {
      result = USBD_HAL_Stall(0);
      if (result == USBD_STATUS_LOCKED) { return ARM_DRIVER_ERROR_BUSY; }
      if (result != USBD_STATUS_SUCCESS) { return ARM_DRIVER_ERROR; }
    }
  } else {  
    result = USBD_HAL_Halt(EP_NUM(ep_addr), stall);
    if (result != stall) return ARM_DRIVER_ERROR; 
  }
  
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_EndpointTransfer (uint8_t ep_addr, uint8_t *data, uint32_t num)
  \brief       Read data from or Write data to USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[out]  data Pointer to buffer for data to read or with data to write
  \param[in]   num  Number of data bytes to transfer
  \return      \ref execution_status
*/
static int32_t USBD_EndpointTransfer (uint8_t ep_addr, uint8_t *data, uint32_t num) {
  if (!(drv_state & USBD_DRIVER_CONNECTED)) { return ARM_DRIVER_ERROR; }
    
  ep[EP_ID(ep_addr)].bufferIndex = 0;
  
  if (USBD_HAL_SetTransferCallback(EP_NUM(ep_addr), (TransferCallback)USBD_TransferCallback, NULL)
      == USBD_STATUS_LOCKED) { return ARM_DRIVER_ERROR_BUSY; }
      
  if (ep_addr&0x80) {
    if (USBD_HAL_Write(EP_NUM(ep_addr), data, num) != USBD_STATUS_SUCCESS) { return ARM_DRIVER_ERROR; }
  } else {
    if (USBD_HAL_Read(EP_NUM(ep_addr), data, num) != USBD_STATUS_SUCCESS) { return ARM_DRIVER_ERROR; }
  }
  return ARM_DRIVER_OK;
}

/**
  \fn          uint32_t USBD_EndpointTransferGetResult (uint8_t ep_addr)
  \brief       Get result of USB Endpoint transfer.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \return      number of successfully transferred data bytes
*/
static uint32_t USBD_EndpointTransferGetResult (uint8_t ep_addr) {
  return ep[EP_ID(ep_addr)].bufferIndex;
}

/**
  \fn          int32_t USBD_EndpointTransferAbort (uint8_t ep_addr)
  \brief       Abort current USB Endpoint transfer.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \return      \ref execution_status
*/
static int32_t USBD_EndpointTransferAbort (uint8_t ep_addr) {
  if (!(drv_state & USBD_DRIVER_CONNECTED)) { return ARM_DRIVER_ERROR; }
    
  UDPHS_EndOfTransfer(ep_addr, USBD_STATUS_ABORTED);
  return ARM_DRIVER_OK;
}

/**
  \fn          uint16_t USBD_GetFrameNumber (void)
  \brief       Get current USB Frame Number.
  \return      Frame Number
*/
static uint16_t USBD_GetFrameNumber (void) {
  return ((USBHS->USBHS_DEVFNUM & USBHS_DEVFNUM_FNUM_Msk) >> USBHS_DEVFNUM_FNUM_Pos);
}


ARM_DRIVER_USBD Driver_USBD0 = {
  USBD_GetVersion,
  USBD_GetCapabilities,
  USBD_Initialize,
  USBD_Uninitialize,
  USBD_PowerControl,
  USBD_DeviceConnect,
  USBD_DeviceDisconnect,
  USBD_DeviceGetState,
  USBD_DeviceRemoteWakeup,
  USBD_DeviceSetAddress,
  USBD_ReadSetupPacket,
  USBD_EndpointConfigure,
  USBD_EndpointUnconfigure,
  USBD_EndpointStall,
  USBD_EndpointTransfer,
  USBD_EndpointTransferGetResult,
  USBD_EndpointTransferAbort,
  USBD_GetFrameNumber
};
